/*
 * Copyright (c) 2004 Chris Wareham.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>

#include "dialogs.h"

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

static void position_dialog(Widget, Widget);

void
confirm_dialog(Widget parent, const char *message,
	Callback yes_callback, Callback no_callback, XtPointer callback_data)
{
	Widget dialog, form, box, label, button;
	int n;
	Arg args[5];

	dialog = XtCreatePopupShell("confirm_dialog", transientShellWidgetClass, parent,
		NULL, 0);

	form = XtCreateManagedWidget("form", formWidgetClass, dialog,
		NULL, 0);

	n = 0;
	XtSetArg(args[n], XtNtop, XtChainTop); n++;
	XtSetArg(args[n], XtNleft, XtChainLeft); n++;
	XtSetArg(args[n], XtNright, XtChainRight); n++;
	XtSetArg(args[n], XtNlabel, message); n++;

	label = XtCreateManagedWidget("confirm_label", labelWidgetClass, form,
		args, n);

	n = 0;
	XtSetArg(args[n], XtNfromVert, label); n++;
	XtSetArg(args[n], XtNleft, XtChainLeft); n++;
	XtSetArg(args[n], XtNright, XtChainRight); n++;
	XtSetArg(args[n], XtNbottom, XtChainBottom); n++;
	XtSetArg(args[n], XtNorientation, XtEhorizontal); n++;

	box = XtCreateManagedWidget("box", boxWidgetClass, form,
		args, n);

	button = XtCreateManagedWidget("yes", commandWidgetClass, box,
		NULL, 0);

	XtAddCallback(button, XtNcallback, yes_callback, callback_data);

	button = XtCreateManagedWidget("no", commandWidgetClass, box,
		NULL, 0);

	XtAddCallback(button, XtNcallback, no_callback, callback_data);

	XtRealizeWidget(dialog);

	position_dialog(dialog, parent);

	XtPopup(dialog, XtGrabExclusive);
	XMapRaised(XtDisplay(dialog), XtWindow(dialog));
}

void
message_dialog(Widget parent, const char *message,
	Callback callback, XtPointer callback_data)
{
	Widget dialog, form, label, button;
	int n;
	Arg args[4];

	dialog = XtCreatePopupShell("message_dialog", transientShellWidgetClass, parent,
		NULL, 0);

	form = XtCreateManagedWidget("form", formWidgetClass, dialog,
		NULL, 0);

	n = 0;
	XtSetArg(args[n], XtNtop, XtChainTop); n++;
	XtSetArg(args[n], XtNleft, XtChainLeft); n++;
	XtSetArg(args[n], XtNright, XtChainRight); n++;
	XtSetArg(args[n], XtNlabel, message); n++;

	label = XtCreateManagedWidget("message_label", labelWidgetClass, form,
		args, n);

	n = 0;
	XtSetArg(args[n], XtNfromVert, label); n++;
	XtSetArg(args[n], XtNleft, XtChainLeft); n++;
	XtSetArg(args[n], XtNbottom, XtChainBottom); n++;

	button = XtCreateManagedWidget("ok", commandWidgetClass, form,
		args, n);

	XtAddCallback(button, XtNcallback, callback, callback_data);

	XtRealizeWidget(dialog);

	position_dialog(dialog, parent);

	XtPopup(dialog, XtGrabExclusive);
	XMapRaised(XtDisplay(dialog), XtWindow(dialog));
}

void
destroy_dialog(Widget widget)
{
	do {
		widget = XtParent(widget);
	} while (XtParent(widget) && XtIsShell(widget) == FALSE);

	XtDestroyWidget(widget);
}

void
position_dialog(Widget dialog, Widget parent)
{
	int displayw, displayh;
	Position dialogx, dialogy, parentx, parenty;
	Dimension dialogw, dialogh, parentw, parenth, border_width;
	int n;
	Arg args[4];

	displayw = DisplayWidth(XtDisplay(parent), DefaultScreen(XtDisplay(parent)));
	displayh = DisplayHeight(XtDisplay(parent), DefaultScreen(XtDisplay(parent)));

	n = 0;
	XtSetArg(args[n], XtNx, &parentx); n++;
	XtSetArg(args[n], XtNy, &parenty); n++;
	XtSetArg(args[n], XtNwidth, &parentw); n++;
	XtSetArg(args[n], XtNheight, &parenth); n++;
	XtGetValues(parent, args, n);

	n = 0;
	XtSetArg(args[n], XtNwidth, &dialogw); n++;
	XtSetArg(args[n], XtNheight, &dialogh); n++;
	XtSetArg(args[n], XtNborderWidth, &border_width); n++;
	XtGetValues(dialog, args, n);

	dialogx = MAX(0, MIN(parentx + (parentw - dialogw) / 2,
		displayw - dialogw - 2 * border_width));

	dialogy = MAX(0, MIN(parenty + (parenth - dialogh) / 2,
		displayh - dialogh - 2 * border_width));

	n = 0;
	XtSetArg(args[n], XtNx, dialogx); n++;
	XtSetArg(args[n], XtNy, dialogy); n++;
	XtSetValues(dialog, args, n);
}
