/*
 * Copyright (c) 2004 Chris Wareham.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/cdio.h>
#include <sys/ioctl.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <util.h>

#include "functions.h"

int
cd_open(int *fd, const char *device)
{
	char buf[64];

	if (*fd > -1)
		close(*fd);

	*fd = opendisk(device, O_RDONLY, buf, sizeof(buf), 0);
	if (*fd < 0) {
		/* ENXIO means "no disc in drive" in this context */
		if (errno == ENXIO) {
			fprintf(stderr, "no disc in drive %s", buf);
			return 0;
		}
		return 1;
	}

	return 0;
}

void
cd_close(int fd)
{
	if (fd > -1)
		close(fd);
}

int
cd_status(int fd, CDStatus *status)
{
	struct ioc_read_subchannel irs;
	struct cd_sub_channel_info data;

	memset(&irs, 0, sizeof(irs));
	irs.data = &data;
	irs.data_len = sizeof(data);
	irs.address_format = CD_MSF_FORMAT;
	irs.data_format = CD_CURRENT_POSITION;

	if (ioctl(fd, CDIOCREADSUBCHANNEL, &irs) < 0) {
		fprintf(stderr, "unable to read subchannel: %s\n", strerror(errno));
		return 1;
	}

	status->status = irs.data->header.audio_status;
	status->track = irs.data->what.position.track_number;
	status->index = irs.data->what.position.index_number;
	status->minute = irs.data->what.position.reladdr.msf.minute;
	status->second = irs.data->what.position.reladdr.msf.second;
	status->frame = irs.data->what.position.reladdr.msf.frame;

	return 0;
}

int
cd_play(int fd, int track)
{
	struct ioc_toc_header th;
	struct ioc_play_track pt;

	if (fd > -1) {
		if (ioctl(fd, CDIOREADTOCHEADER, &th) < 0) {
			fprintf(stderr, "unable to read TOC header: %s\n", strerror(errno));
			return 1;
		}

		pt.start_track = track;
		pt.start_index = 1;
		pt.end_track = th.ending_track;
		pt.end_index = 99;

		if (ioctl(fd, CDIOCPLAYTRACKS, &pt) < 0) {
			fprintf(stderr, "unable to read subchannel: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}

int
cd_pause(int fd)
{
	if (fd > -1) {
		if (ioctl(fd, CDIOCPAUSE) < 0) {
			fprintf(stderr, "unable to pause: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}

int
cd_resume(int fd)
{
	if (fd > -1) {
		if (ioctl(fd, CDIOCRESUME) < 0) {
			fprintf(stderr, "unable to resume: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}

int
cd_skip(int fd, int direction)
{
	struct ioc_toc_header th;
	CDStatus status;

	if (fd < 0)
		return 1;

	if (ioctl(fd, CDIOREADTOCHEADER, &th) < 0) {
		fprintf(stderr, "unable to read TOC header: %s\n", strerror(errno));
		return 1;
	}

	if (cd_status(fd, &status))
		return 1;

	status.track += direction;
	if (status.track > th.ending_track)
		status.track = th.starting_track;
	else if(status.track < th.starting_track)
		status.track = th.ending_track;

	return cd_play(fd, status.track);
}

int
cd_stop(int fd)
{
	if (fd > -1) {
		if (ioctl(fd, CDIOCSTOP) < 0) {
			fprintf(stderr, "unable to stop: %s\n", strerror(errno));
			return 1;
		}

		if (ioctl(fd, CDIOCALLOW) < 0) {
			fprintf(stderr, "unable to allow: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}

int
cd_eject(int fd)
{
	if (fd > -1) {
		if (ioctl(fd, CDIOCALLOW) < 0) {
			fprintf(stderr, "unable to allow: %s\n", strerror(errno));
			return 1;
		}
		if (ioctl(fd, CDIOCEJECT) < 0) {
			fprintf(stderr, "unable to eject: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}

int
cd_get_volume(int fd, int *volume)
{
	struct ioc_vol iv;

	*volume = 0;

	if (fd > -1) {
		if (ioctl(fd, CDIOCGETVOL, &iv) < 0) {
			fprintf(stderr, "unable to get volume: %s\n", strerror(errno));
			return 1;
		}

		*volume = iv.vol[0] > iv.vol[1] ? iv.vol[0] : iv.vol[1];
	}

	return 0;
}

int
cd_set_volume(int fd, int volume)
{
	struct ioc_vol iv;

	if (fd > -1) {
		iv.vol[0] = volume;
		iv.vol[1] = volume;
		iv.vol[2] = 0;
		iv.vol[3] = 0;

		if (ioctl(fd, CDIOCSETVOL, &iv) < 0) {
			fprintf(stderr, "unable to set volume: %s\n", strerror(errno));
			return 1;
		}
	}

	return 0;
}
