#include <sys/types.h>
#include <sys/cdio.h>
#include <sys/ioctl.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <util.h>

#define CD_MAX_TRACK 99

static int info(void);
static int read_toc_entrys(int);
static void print_toc_entry(struct cd_toc_entry *);
static void lba2msf(unsigned long, unsigned *, unsigned *, unsigned *);
static unsigned msf2lba(unsigned, unsigned, unsigned);

static int fd;
static struct cd_toc_entry toc_buffer[CD_MAX_TRACK + 1];

int
main(int argc, char *argv[])
{
	char device[64], buf[64];

	snprintf(device, sizeof(device), "cd0%c", 'a' + getrawpartition());

	if ((fd = opendisk(device, O_RDONLY, buf, sizeof(buf), 0)) < 0) {
		/* ENXIO means "no disc in drive" in this context */
		if (errno == ENXIO) {
			fprintf(stderr, "no disc in drive %s", buf);
			return 0;
		}
		return 1;
	}

	info();

	close(fd);

	return 0;
}

int
info(void)
{
	struct ioc_toc_header h;
	int n, i;

	if (ioctl(fd, CDIOREADTOCHEADER, &h) < 0) {
		fprintf(stderr, "unable to read TOC header: %s", strerror(errno));
		return 1;
	}

	n = h.ending_track - h.starting_track + 1;

	if (read_toc_entrys((n + 1) * sizeof(struct cd_toc_entry)) < 0)
		return 1;

	printf("track     start  duration\n");
	printf("-------------------------\n");

	for (i = 0; i < n; i++) {
		printf("%5d  ", toc_buffer[i].track);
		print_toc_entry(toc_buffer + i);
	}
	/* Lead-out area */
	printf("    -  ");
	print_toc_entry(toc_buffer + n);

	return 0;
}

int
read_toc_entrys(int len)
{
	struct ioc_read_toc_entry t;

	t.address_format = CD_MSF_FORMAT;
	t.starting_track = 0;
	t.data_len = len;
	t.data = toc_buffer;

	if (ioctl(fd, CDIOREADTOCENTRYS, &t) < 0) {
		fprintf(stderr, "unable to read TOC entrys: %s", strerror(errno));
		return 1;
	}

	return 0;
}

void
print_toc_entry(struct cd_toc_entry *e)
{
	int block, next;
	unsigned m, s, f;

	/* Print track start */
	printf("%2d:%02d.%02d  ", e->addr.msf.minute, e->addr.msf.second,
		e->addr.msf.frame);

	if (e->track > CD_MAX_TRACK) {
		/* Lead-out area */
		printf("       -\n");
		return;
	}

	block = msf2lba(e->addr.msf.minute, e->addr.msf.second, e->addr.msf.frame);
	next = msf2lba(e[1].addr.msf.minute, e[1].addr.msf.second, e[1].addr.msf.frame);
	lba2msf(next - block, &m, &s, &f);

	/* Print track duration */
	printf("%2d:%02d.%02d\n", m, s, f);
}

void
lba2msf(unsigned long lba, unsigned *m, unsigned *s, unsigned *f)
{
	lba += 150; /* block start offset */
	lba &= 0xffffff; /* negative lbas use only 24 bits */
	*m = lba / (60 * 75);
	lba %= (60 * 75);
	*s = lba / 75;
	*f = lba % 75;
}

/*
     1 frame equals 1 block
     75 frames equals one second
     2 second start offset (150 blocks)
   So to determine LBA from MSF:
     ((total seconds * 75) + frames) - 150
*/

unsigned
msf2lba(unsigned m, unsigned s, unsigned f)
{
	return (((m * 60) + s) * 75 + f) - 150;
}
