/*
 * Copyright (c) 2004 Chris Wareham.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/cdio.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <util.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>

#include "Gridbox.h"
#include "Slider.h"

#include "bitmaps/eject.xbm"
#include "bitmaps/icon.xbm"
#include "bitmaps/pause.xbm"
#include "bitmaps/play.xbm"
#include "bitmaps/scanb.xbm"
#include "bitmaps/scanf.xbm"
#include "bitmaps/stop.xbm"
#include "bitmaps/trackb.xbm"
#include "bitmaps/trackf.xbm"

#include "dialogs.h"
#include "functions.h"
#include "xawcd.h"

static void usage(const char *);
static void create_interface(Interface *);
static void create_lcd_pixmaps(Interface *);
static void show_interface(Interface *);
static void close_cb(Widget, XtPointer, XtPointer);
static void error_cb(Widget, XtPointer, XtPointer);
static void about_cb(Widget, XtPointer, XtPointer);
static void play_cb(Widget, XtPointer, XtPointer);
static void pause_cb(Widget, XtPointer, XtPointer);
static void stop_cb(Widget, XtPointer, XtPointer);
static void eject_cb(Widget, XtPointer, XtPointer);
static void trackb_cb(Widget, XtPointer, XtPointer);
static void trackf_cb(Widget, XtPointer, XtPointer);
static void quit_cb(Widget, XtPointer, XtPointer);
static void quityes_cb(Widget, XtPointer, XtPointer);
static void quitno_cb(Widget, XtPointer, XtPointer);
static void volume_cb(Widget, XtPointer, XtPointer);
static void timeout(XtPointer, XtIntervalId *);

static char *defaults[] = {
#include "XawCD.h"
NULL
};

int
main(int argc, char *argv[])
{
	int c;
	char *progname, *ptr, device[64];
	Interface ui;

	progname = argv[0];
	if ((ptr = strrchr(progname, '/')))
		progname = ptr + 1;

	device[0] = '\0';

	while ((c = getopt(argc, argv, "f:h")) != -1) {
		switch (c) {
		case 'f':
			snprintf(device, sizeof(device), "%s", optarg);
			continue;
		case 'h':
		default:
			usage(progname);
		}
	}
	argc -= optind;
	argv += optind;

	if (device[0] == '\0')
		snprintf(device, sizeof(device), "cd0%c", 'a' + getrawpartition());

	argv[0] = "xawcd";

	XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

	ui.toplevel = XtAppInitialize(&ui.app_context, "XawCD",
		NULL, 0, &argc, argv, defaults, NULL, 0);

	if (cd_open(&ui.fd, device)) {
		message_dialog(ui.toplevel, "Unable to open CD device", error_cb, NULL);
		XtAppMainLoop(ui.app_context);
	}

	create_interface(&ui);

	show_interface(&ui);

	XtAppMainLoop(ui.app_context);

	return 0;
}

void
usage(const char *progname)
{
	fprintf(stderr, "usage: %s [-f device] [-h]\n", progname);
	exit(1);
}

void
create_interface(Interface *ui)
{
	Pixmap pixmap;
	Widget gridbox, box, button, lcd, lcd_box;
	char buf[64];
	int volume;

	/* set an icon if one isn't set in the resources */

	XtVaGetValues(ui->toplevel, XtNiconPixmap, &pixmap, NULL);

	if (pixmap == None) {
		pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
			RootWindowOfScreen(XtScreen(ui->toplevel)),
			icon_bits, icon_width, icon_height);
		XtVaSetValues(ui->toplevel, XtNiconPixmap, pixmap, NULL);
	}

	/* gridbox layout */

	gridbox = XtVaCreateManagedWidget("gridbox", gridboxWidgetClass, ui->toplevel,
		XtNtop,    XtChainTop,
		XtNleft,   XtChainLeft,
		XtNright,  XtChainRight,
		XtNbottom, XtChainBottom,
		NULL);

	/* version label */

	snprintf(buf, sizeof(buf), "XawCD v%d.%d", MAJOR_VERSION, MINOR_VERSION);

	XtVaCreateManagedWidget("version_label", labelWidgetClass, gridbox,
		XtNlabel,     buf,
		XtNgridx,     0,
		XtNgridy,     0,
		XtNweightx,   1,
		XtNweighty,   0,
		XtNgridWidth, 4,
		XtNfill,      FillWidth,
		XtNgravity,   CenterGravity,
		NULL);

	/* LCD label */

	create_lcd_pixmaps(ui);

	lcd = XtVaCreateManagedWidget("lcd", gridboxWidgetClass, gridbox,
		XtNgridx,     0,
		XtNgridy,     1,
		XtNweightx,   1,
		XtNweighty,   0,
		XtNgridWidth, 4,
		XtNfill,      FillWidth,
		NULL);

	lcd_box = XtVaCreateManagedWidget("lcd_box", boxWidgetClass, lcd,
		XtNgridx,     0,
		XtNgridy,     0,
		XtNweightx,   1,
		XtNweighty,   0,
		XtNfill,      FillNone,
		XtNgravity,   CenterGravity,
		XtNorientation, XtEhorizontal,
		NULL);

	ui->lcd_logo = XtVaCreateManagedWidget("lcd_logo", labelWidgetClass, lcd_box,
		XtNbitmap, ui->stop_pixmap,
		NULL);

	ui->lcd_label = XtVaCreateManagedWidget("lcd_label", labelWidgetClass, lcd_box,
		XtNlabel, "00 00:00",
		NULL);

	/* play button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		play_bits, play_width, play_height);

	ui->play_button = XtVaCreateManagedWidget("play", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   0,
		XtNgridy,   2,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->play_button, XtNcallback, play_cb, ui);

	/* pause button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		pause_bits, pause_width, pause_height);

	ui->pause_button = XtVaCreateManagedWidget("pause", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   1,
		XtNgridy,   2,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->pause_button, XtNcallback, pause_cb, ui);

	/* stop button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		stop_bits, stop_width, stop_height);

	ui->stop_button = XtVaCreateManagedWidget("stop", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   2,
		XtNgridy,   2,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->stop_button, XtNcallback, stop_cb, ui);

	/* eject button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		eject_bits, eject_width, eject_height);

	ui->eject_button = XtVaCreateManagedWidget("eject", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   3,
		XtNgridy,   2,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->eject_button, XtNcallback, eject_cb, ui);

	/* track back button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		trackb_bits, trackb_width, trackb_height);

	ui->trackb_button = XtVaCreateManagedWidget("trackb", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   0,
		XtNgridy,   3,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->trackb_button, XtNcallback, trackb_cb, ui);

	/* scan back button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		scanb_bits, scanb_width, scanb_height);

	ui->scanb_button = XtVaCreateManagedWidget("scanb", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   1,
		XtNgridy,   3,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

/*	XtAddCallback(ui->scanb_button, XtNcallback, scanb_cb, ui); */

	/* scan forward button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		scanf_bits, scanf_width, scanf_height);

	ui->scanf_button = XtVaCreateManagedWidget("scanf", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   2,
		XtNgridy,   3,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

/*	XtAddCallback(ui->scanf_button, XtNcallback, scanf_cb, ui); */

	/* track forward button */

	pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		trackf_bits, trackf_width, trackf_height);

	ui->trackf_button = XtVaCreateManagedWidget("trackf", commandWidgetClass, gridbox,
		XtNbitmap,  pixmap,
		XtNgridx,   3,
		XtNgridy,   3,
		XtNweightx, 1,
		XtNweighty, 1,
		XtNfill,    FillBoth,
		NULL);

	XtAddCallback(ui->trackf_button, XtNcallback, trackf_cb, ui);

	/* volume slider */

	cd_get_volume(ui->fd, &volume);

	ui->volume = XtVaCreateManagedWidget("volume", sliderWidgetClass, gridbox,
		XtNvalue,     volume,
		XtNgridx,     0,
		XtNgridy,     4,
		XtNweightx,   1,
		XtNweighty,   0,
		XtNgridWidth, 4,
		XtNfill,      FillWidth,
		NULL);

	XtAddCallback(ui->volume, XtNvalueChangedCallback, volume_cb, ui);

	/* about and quit buttons */

	box = XtVaCreateManagedWidget("box", boxWidgetClass, gridbox,
		XtNgridx,     0,
		XtNgridy,     5,
		XtNweightx,   0,
		XtNweighty,   0,
		XtNgridWidth, 4,
		XtNfill,      FillNone,
		XtNgravity,   CenterGravity,
		XtNorientation, XtEhorizontal,
		NULL);

	button = XtCreateManagedWidget("about", commandWidgetClass, box,
		NULL, 0);

	XtAddCallback(button, XtNcallback, about_cb, ui);

	button = XtCreateManagedWidget("quit", commandWidgetClass, box,
		NULL, 0);

	XtAddCallback(button, XtNcallback, quit_cb, ui);
}

void
create_lcd_pixmaps(Interface *ui)
{
	ui->stop_pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		stop_bits, stop_width, stop_height);

	ui->play_pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		play_bits, play_width, play_height);

	ui->pause_pixmap = XCreateBitmapFromData(XtDisplay(ui->toplevel),
		RootWindowOfScreen(XtScreen(ui->toplevel)),
		pause_bits, pause_width, pause_height);
}

void
show_interface(Interface *ui)
{
	XtRealizeWidget(ui->toplevel);

	ui->state = STATE_STOP;
	ui->timeout = 0;
	ui->timer = XtAppAddTimeOut(ui->app_context, TIMEOUT, timeout, ui);
}

void
close_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	destroy_dialog(widget);
}

void
error_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	destroy_dialog(widget);
	exit(1);
}

void
about_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;

	ui = (Interface *)client_data;
	message_dialog(ui->toplevel, "XawCD (C) Chris Wareham 2004", close_cb, NULL);
}

void
play_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	CDStatus status;
	int track = 1;

	ui = (Interface *)client_data;

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
		case CD_AS_PLAY_PAUSED:
			track = status.track;
			break;
		default:
			track = 1;
			break;
		}
	}

	cd_play(ui->fd, track);
}

void
pause_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	CDStatus status;

	ui = (Interface *)client_data;

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
			cd_pause(ui->fd);
			break;
		case CD_AS_PLAY_PAUSED:
			cd_resume(ui->fd);
			break;
		}
	}
}

void
stop_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	CDStatus status;

	ui = (Interface *)client_data;

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
		case CD_AS_PLAY_PAUSED:
			cd_stop(ui->fd);
			break;
		}
	}
}

void
eject_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	CDStatus status;

	ui = (Interface *)client_data;

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
		case CD_AS_PLAY_PAUSED:
			cd_stop(ui->fd);
			break;
		}
	}

	cd_eject(ui->fd);
}

void
trackb_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;

	ui = (Interface *)client_data;

	cd_skip(ui->fd, -1);
}

void
trackf_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;

	ui = (Interface *)client_data;

	cd_skip(ui->fd, 1);
}

void
quit_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;

	ui = (Interface *)client_data;

	confirm_dialog(ui->toplevel, "Are you sure you want to quit?",
		quityes_cb, quitno_cb, ui);
}

void
quityes_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	CDStatus status;

	ui = (Interface *)client_data;

	destroy_dialog(widget);

	XtRemoveTimeOut(ui->timer);

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
		case CD_AS_PLAY_PAUSED:
			cd_stop(ui->fd);
			break;
		}
	}
	cd_close(ui->fd);

	XtUnmapWidget(ui->toplevel);
	XtDestroyApplicationContext(ui->app_context);

	exit(0);
}

void
quitno_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	destroy_dialog(widget);
}

void
volume_cb(Widget widget, XtPointer client_data, XtPointer call_data)
{
	Interface *ui;
	int volume;

	ui = (Interface *)client_data;

	XtVaGetValues(ui->volume, XtNvalue, &volume, NULL);

	cd_set_volume(ui->fd, volume);
}

void
timeout(XtPointer client_data, XtIntervalId *timer)
{
	Interface *ui;
	CDStatus status;
	char buf[9];

	ui = (Interface *)client_data;

	ui->timeout++;

	if (cd_status(ui->fd, &status) == 0) {
		switch (status.status) {
		case CD_AS_PLAY_IN_PROGRESS:
			if (ui->state != STATE_PLAY || ui->timeout % 4 == 0) {
				snprintf(buf, sizeof(buf), "%02d %02d:%02d",
					status.track, status.minute, status.second);
				XtVaSetValues(ui->lcd_logo, XtNbitmap, ui->play_pixmap, NULL);
				XtVaSetValues(ui->lcd_label, XtNlabel, buf, NULL);
				ui->state = STATE_PLAY;
				ui->timeout = 0;
			}
			break;
		case CD_AS_PLAY_PAUSED:
			if (ui->state != STATE_PAUSE) {
				snprintf(buf, sizeof(buf), "%02d %02d:%02d",
					status.track, status.minute, status.second);
				XtVaSetValues(ui->lcd_logo, XtNbitmap, ui->pause_pixmap, NULL);
				XtVaSetValues(ui->lcd_label, XtNlabel, buf, NULL);
				ui->state = STATE_PAUSE;
			}
			break;
		default:
			if (ui->state != STATE_STOP) {
				XtVaSetValues(ui->lcd_logo, XtNbitmap, ui->stop_pixmap, NULL);
				XtVaSetValues(ui->lcd_label, XtNlabel, "00 00:00", NULL);
				ui->state = STATE_STOP;
			}
			break;
		}
	}

	ui->timer = XtAppAddTimeOut(ui->app_context, TIMEOUT, timeout, ui);
}
